/**
 * Copyright 2011 J Adam Crain (jadamcrain@gmail.com)
 *
 * Licensed to J Adam Crain under one or more contributor license agreements.
 * See the NOTICE file distributed with this work for additional information
 * regarding copyright ownership. J Adam Crain licenses this file to you under
 * the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License
 * at:
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.automatak.nio4s

import java.nio.channels._
import java.nio.ByteBuffer
import com.automatak.executor4s._
import java.net.SocketAddress
import java.nio.file.Files
import java.util.concurrent.Executor
import com.automatak.executor4s.Success

package object dsl {

  implicit def decorateAsynchronousByteChannel(channel: AsynchronousByteChannel) =
    new AsynchronousByteChannelDecorator(channel)

  implicit def decorateAsynchronousServerSocketChannel(channel: AsynchronousServerSocketChannel) =
    new AsynchronousServerSocketChannelDecorator(channel)

  implicit def decorateAsynchronousSocketChannel(channel: AsynchronousSocketChannel) =
    new AsynchronousSocketChannelDecorator(channel)

  implicit def decorateAsynchronousFileChannel(channel: AsynchronousFileChannel) =
    new AsynchronousFileChannelDecorator(channel)
}

object ExceptionHandler {

  def reportExceptions[A](callback: Result[A] => Unit)(fun: => Unit): Unit = {
    try {
      fun
    } catch {
      case ex: Exception => callback(Failure(ex))
    }
  }
}

class IntegerCompletionHandler(callback: Result[Int] => Unit) extends CompletionHandler[java.lang.Integer, Void] {
  def completed(result: java.lang.Integer, attachment: Void) = callback(Success(result))
  def failed(exc: Throwable, attachment: Void) = callback(Failure(exc))
}

class AsynchronousFileChannelDecorator(channel: AsynchronousFileChannel) {
  import ExceptionHandler._

  def writeAll(src: ByteBuffer, position: Long)(implicit exe: Executor): Future[ByteBuffer] = {
    val f = Future.pending[ByteBuffer]
    writeAsyncAll(src, position)(f.set)
    f
  }

  def readAll(src: ByteBuffer, position: Long)(implicit exe: Executor): Future[ByteBuffer] = {
    val f = Future.pending[ByteBuffer]
    readAsyncAll(src, position)(f.set)
    f
  }

  def writeAsyncOnce(src: ByteBuffer, position: Long)(callback: Result[Int] => Unit): Unit = {
    reportExceptions(callback)(channel.write(src, position, null, new IntegerCompletionHandler(callback)))
  }

  def readAsyncOnce(dst: ByteBuffer, position: Long)(callback: Result[Int] => Unit): Unit = {
    reportExceptions(callback)(channel.read(dst, position, null, new IntegerCompletionHandler(callback)))
  }

  def writeAsyncAll(src: ByteBuffer, position: Long)(callback: Result[ByteBuffer] => Unit): Unit = {
    def onWrite(pos: Long)(res: Result[Int]): Unit = res match {
      case Success(x) =>
        val newpos = pos + x
        if (src.remaining() > 0) writeAsyncOnce(src, newpos)(onWrite(newpos))
        else callback(Success(src))
      case f: Failure => callback(f)
    }
    writeAsyncOnce(src, position)(onWrite(position))
  }

  def readAsyncAll(src: ByteBuffer, position: Long)(callback: Result[ByteBuffer] => Unit): Unit = {
    require(src.remaining() > 0)
    def onRead(pos: Long)(res: Result[Int]): Unit = res match {
      case Success(num) =>
        if (num <= 0) callback(Failure("EOS"))
        else if (src.remaining() > 0) {
          val newpos = pos + num
          readAsyncOnce(src, newpos)(onRead(newpos))
        } else callback(Success(src))
      case f: Failure => callback(f)
    }
    readAsyncOnce(src, position)(onRead(position))
  }
}

class AsynchronousServerSocketChannelDecorator(channel: AsynchronousServerSocketChannel) {

  import ExceptionHandler._

  def acceptAsync(callback: Result[AsynchronousSocketChannel] => Unit): Unit = {
    val handler = new CompletionHandler[AsynchronousSocketChannel, Void] {
      def completed(result: AsynchronousSocketChannel, attachment: Void) = callback(Success(result))
      def failed(exc: Throwable, attachment: Void) = callback(Failure(exc))
    }
    reportExceptions(callback)(channel.accept(null, handler))
  }

}

class AsynchronousByteChannelDecorator(channel: AsynchronousByteChannel) {

  import ExceptionHandler._

  def writeAsyncOnce(src: ByteBuffer)(callback: Result[Int] => Unit): Unit = {
    val handler = new IntegerCompletionHandler(callback)
    reportExceptions(callback)(channel.write(src, null, handler))
  }

  def readAsyncOnce(dst: ByteBuffer)(callback: Result[Int] => Unit): Unit = {
    val handler = new IntegerCompletionHandler(callback)
    reportExceptions(callback)(channel.read(dst, null, handler))
  }

  def writeAsyncAll(src: ByteBuffer)(callback: Result[ByteBuffer] => Unit): Unit = {
    def onWrite(res: Result[Int]): Unit = res match {
      case Success(x) =>
        if (src.remaining() > 0) writeAsyncOnce(src)(onWrite)
        else callback(Success(src))
      case f: Failure => callback(f)
    }
    writeAsyncOnce(src)(onWrite)
  }

  def readAsyncAll(src: ByteBuffer)(callback: Result[ByteBuffer] => Unit): Unit = {
    require(src.remaining() > 0)
    def onRead(res: Result[Int]): Unit = res match {
      case Success(num) =>
        if (num <= 0) callback(Failure("EOS"))
        else if (src.remaining() > 0) readAsyncOnce(src)(onRead)
        else callback(Success(src))
      case f: Failure => callback(f)
    }
    readAsyncOnce(src)(onRead)
  }

}

class AsynchronousSocketChannelDecorator(channel: AsynchronousSocketChannel) {

  import ExceptionHandler._

  def connectAsync(address: SocketAddress)(callback: Result[AsynchronousSocketChannel] => Unit): Unit = {
    val handler = new CompletionHandler[Void, Void] {
      def completed(result: Void, attachment: Void) = callback(Success(channel))
      def failed(exc: Throwable, attachment: Void) = callback(Failure(exc))
    }
    reportExceptions(callback)(channel.connect(address, null, handler))
  }

  def writeAsyncOnceWithTimeout(src: ByteBuffer, timeout: TimeInterval)(callback: Result[Int] => Unit): Unit = {
    val handler = new IntegerCompletionHandler(callback)
    reportExceptions(callback)(channel.write(src, timeout.count, timeout.timeunit, null, handler))
  }

  def readAsyncOnceWithTimeout(dst: ByteBuffer, timeout: TimeInterval)(callback: Result[Int] => Unit): Unit = {
    val handler = new IntegerCompletionHandler(callback)
    reportExceptions(callback)(channel.read(dst, timeout.count, timeout.timeunit, null, handler))
  }

  def writeAsyncAllWithTimeout(src: ByteBuffer, timeout: TimeInterval)(callback: Result[ByteBuffer] => Unit): Unit = {
    def onWrite(res: Result[Int]): Unit = res match {
      case Success(x) =>
        if (src.remaining() > 0) writeAsyncOnceWithTimeout(src, timeout)(onWrite)
        else callback(Success(src))
      case f: Failure => callback(f)
    }
    writeAsyncOnceWithTimeout(src, timeout)(onWrite)
  }

  def readAsyncAllWithTimeout(src: ByteBuffer, timeout: TimeInterval)(callback: Result[ByteBuffer] => Unit): Unit = {
    require(src.remaining() > 0)
    def onRead(res: Result[Int]): Unit = res match {
      case Success(num) =>
        if (num <= 0) callback(Failure("EOS"))
        else if (src.remaining() > 0) readAsyncOnceWithTimeout(src, timeout)(onRead)
        else callback(Success(src))
      case f: Failure => callback(f)
    }
    readAsyncOnceWithTimeout(src, timeout)(onRead)
  }

}

